#include "OdbcServer.h"
#include <iostream.h>
#include <sql.h>
#include <sqlext.h>

QString OdbcServer::sql_error( SQLHANDLE henv, SQLHANDLE hdbc, SQLHANDLE hstmt )
{
    SQLCHAR msg[ SQL_MAX_MESSAGE_LENGTH ];
    SQLINTEGER native;
    SQLSMALLINT len;
    SQLRETURN ret;
    SQLCHAR state[ 6 ];
    QString result;

    ret = SQLError( henv, hdbc, hstmt, state, &native, msg, sizeof( msg ), &len );

    if ( SQL_SUCCEEDED( ret ))
    {
        result.sprintf( "Error:%s:%d:%s", state, native, msg );
    }
    else
    {
        result = "Error: No Error text available";
    }

    return result;
}

OdbcServer::OdbcServer(void)
	:	connected(false)
	{
        // Allocate a env handle
        SQLAllocEnv( &henv ); 
	}

OdbcServer::~OdbcServer(void)
  {
      // Release the env handle
      SQLFreeEnv( henv );
  }

QueryResult OdbcServer::connect(QString host, QString user, QString passwd, int port)
{
    SQLRETURN ret;
    QueryResult res;

    odbc_dsn=host;
    odbc_user=user;
    odbc_passwd=passwd;

    // Allocate a connection handle

    ret = SQLAllocConnect( henv, &hdbc );

    if ( !SQL_SUCCEEDED( ret ))
    {
        res.type = QueryResult::Error;
        res.message = sql_error(henv,SQL_NULL_HDBC,SQL_NULL_HSTMT);
        connected = false;
        return res;
    }

    ret = SQLConnect( hdbc, 
            (SQLCHAR*)(const char *)odbc_dsn, 
            SQL_NTS,
            (SQLCHAR*)(const char *)odbc_user, 
            SQL_NTS,
            (SQLCHAR*)(const char *)odbc_passwd,
            SQL_NTS );

    if ( !SQL_SUCCEEDED( ret ))
    {
        res.type = QueryResult::Error;
        res.message = sql_error(SQL_NULL_HENV,hdbc,SQL_NULL_HSTMT);
        connected = false;
        return res;
    }
    else
    {
        res.type = QueryResult::Ok;
        connected = true;
        return res;
    }
}

void OdbcServer::close(void)
{
    if ( connected )
    {
        SQLDisconnect( hdbc );
    }

    SQLFreeConnect( hdbc );
    connected = false;
}

QueryResult OdbcServer::selectBase(QString baseName)
{
    QueryResult res;

  	res.type = QueryResult::Ok;

  	return res;
}

QueryResult OdbcServer::query(QString query)
  {
    QueryResult res;
    SQLRETURN ret;
    SQLSMALLINT cols;
    SQLHANDLE hstmt;

    ret = SQLAllocStmt( hdbc, &hstmt );
    if ( !SQL_SUCCEEDED( ret ))
    {
        res.type = QueryResult::Error;
        res.message = sql_error(SQL_NULL_HENV,hdbc,SQL_NULL_HSTMT);
        return res;
    }
    
    ret = SQLExecDirect( hstmt, (SQLCHAR*)(const char *)query, SQL_NTS );

    if ( !SQL_SUCCEEDED( ret ))
    {
        res.type = QueryResult::Error;
        res.message = sql_error(SQL_NULL_HENV,SQL_NULL_HDBC,hstmt);
        SQLFreeStmt(hstmt,SQL_DROP);
        return res;
    }

    ret = SQLNumResultCols( hstmt, &cols );
    if ( !SQL_SUCCEEDED( ret ))
    {
        res.type = QueryResult::Error;
        res.message = sql_error(SQL_NULL_HENV,SQL_NULL_HDBC,hstmt);
        SQLFreeStmt(hstmt,SQL_DROP);
        return res;
    }

    if ( cols > 0 )
    {
        res.type = QueryResult::Set;
        if ( !putTuples( res,  hstmt ))
        {
            res.type = QueryResult::Error;
            res.message = sql_error(SQL_NULL_HENV,SQL_NULL_HDBC,hstmt);
            SQLFreeStmt(hstmt,SQL_DROP);
            return res;
        }
    }
    else
    {
        res.type = QueryResult::Ok;
    }

    SQLFreeStmt(hstmt,SQL_DROP);

    return res;
}
	
bool OdbcServer::putTuples( QueryResult &res, SQLHANDLE hstmt )
{
    SQLRETURN ret;
    SQLSMALLINT cols;
    int row_count = 0;

    ret = SQLNumResultCols( hstmt, &cols );
    if ( !SQL_SUCCEEDED( ret ))
    {
        return false;
    }

	res.numColumns = cols;
    res.fields = "";

    for( int i = 1; i < cols + 1; i ++ )
    {
        SQLCHAR name[ 128 ];
        SQLSMALLINT len_name;
        SQLSMALLINT type;

        ret = SQLDescribeCol( hstmt, i, 
                name, sizeof( name ), &len_name,
                &type, NULL, NULL, NULL );

        if ( !SQL_SUCCEEDED( ret ))
        {
            return false;
        }

		res.fields += ( char * ) name;
		if( i < cols ) 
        {
			res.fields += '\t';
        }
    }

	res.result.clear();

    while( SQL_SUCCEEDED( ret = SQLFetch( hstmt )))
    {
		QString tuple = "";

        for( int i = 1; i < cols + 1; i ++ )
        {
            SQLCHAR value[ 128 ];
            SQLINTEGER value_len;

            ret = SQLGetData( hstmt, i, SQL_C_CHAR, 
                    value, sizeof( value ), &value_len );

            if ( !SQL_SUCCEEDED( ret ))
            {
                return false;
            }

            if ( value_len != SQL_NULL_DATA )
            {
			    tuple += ( char * ) value;
            }
            else
            {
                tuple += "<NULL>";
            }

            if( i < cols ) 
            {
                tuple += '\t';
            }
        }

		res.result.append(tuple);
    }

    if ( ret != SQL_NO_DATA )
    {
        return false;
    }

    res.type = QueryResult::Set;
    res.message = "Query OK";

    return true;
}

QueryResult OdbcServer::getBases(void)
{
    QueryResult res;
    SQLRETURN ret;
    SQLHANDLE hstmt;

    res.type = QueryResult::Set;
    res.result.append( odbc_dsn );
    return res;

    // obtain a list of schema

    ret = SQLAllocStmt( hdbc, &hstmt );
    if ( !SQL_SUCCEEDED( ret ))
    {
        res.type = QueryResult::Error;
        res.message = sql_error(SQL_NULL_HENV,hdbc,SQL_NULL_HSTMT);
        return res;
    }
    
    ret = SQLTables( hstmt, 
            (SQLCHAR*)"", 0, 
            (SQLCHAR*)SQL_ALL_SCHEMAS, SQL_NTS, 
            (SQLCHAR*)"", 0, 
            (SQLCHAR*)"", 0 );

    if ( !SQL_SUCCEEDED( ret ))
    {
        res.type = QueryResult::Error;
        res.message = sql_error(SQL_NULL_HENV,SQL_NULL_HDBC,hstmt);
        SQLFreeStmt(hstmt,SQL_DROP);
        return res;
    }

    res.type = QueryResult::Set;

    while( SQL_SUCCEEDED(( ret = SQLFetch( hstmt ))))
    {
        char schema[ 128 ];
        SQLINTEGER len;
        SQLRETURN ret1;

        ret1 = SQLGetData( hstmt, 2, SQL_C_CHAR, schema, sizeof( schema ),
                &len );

        if ( SQL_SUCCEEDED( ret1 ) && len != SQL_NULL_DATA )
        {
  	        res.result.append( schema );
        }
    }

    if ( ret != SQL_NO_DATA )
    {
        res.type = QueryResult::Error;
        res.message = sql_error(SQL_NULL_HENV,SQL_NULL_HDBC,hstmt);
        SQLFreeStmt(hstmt,SQL_DROP);
        return res;
    }

    SQLFreeStmt(hstmt,SQL_DROP);

    return res;
}

QueryResult OdbcServer::getTables(QString baseName, QString tableName)
{
    QueryResult res;
    SQLCHAR *ptables;
    SQLRETURN ret;
    SQLHANDLE hstmt;

    if ( tableName.isNull() )
    {
      ptables = NULL;
    }
    else
    {
      ptables = (SQLCHAR*)(const char *) tableName;
    }

    res.type = QueryResult::Set;

    // obtain a list of schema

    ret = SQLAllocStmt( hdbc, &hstmt );
    if ( !SQL_SUCCEEDED( ret ))
    {
        res.type = QueryResult::Error;
        res.message = sql_error(SQL_NULL_HENV,hdbc,SQL_NULL_HSTMT);
        return res;
    }
    
    ret = SQLTables( hstmt, 
            (SQLCHAR*)"", 0, 
            (SQLCHAR*)"", 0, 
            ptables, 0, 
            (SQLCHAR*)"", 0 );

    if ( !SQL_SUCCEEDED( ret ))
    {
        res.type = QueryResult::Error;
        res.message = sql_error(SQL_NULL_HENV,SQL_NULL_HDBC,hstmt);
        SQLFreeStmt(hstmt,SQL_DROP);
        return res;
    }

    res.type = QueryResult::Set;

    while( SQL_SUCCEEDED(( ret = SQLFetch( hstmt ))))
    {
        char catalog[ 128 ], schema[ 128 ], table[ 128 ];
        SQLINTEGER len1, len2, len3;
        SQLRETURN ret1, ret2, ret3;

        ret1 = SQLGetData( hstmt, 1, SQL_C_CHAR, catalog, sizeof( catalog ),
                &len1 );
        ret2 = SQLGetData( hstmt, 2, SQL_C_CHAR, schema, sizeof( schema ),
                &len2 );
        ret3 = SQLGetData( hstmt, 3, SQL_C_CHAR, table, sizeof( table ),
                &len3 );

        if ( SQL_SUCCEEDED( ret3 ) && len3 != SQL_NULL_DATA )
        {
            char name[ 512 ];

            name[ 0 ] = '\0';

            if ( len1 > 0 )
            {
                strcat( name, catalog );
            }
            if ( len2 > 0 )
            {
                if ( strlen( name ))
                    strcat( name, "." );
                strcat( name, schema );
            }
            if ( strlen( name ))
                strcat( name, "." );

            strcat( name, table );
  	        res.result.append( name );
        }
    }

    if ( ret != SQL_NO_DATA )
    {
        res.type = QueryResult::Error;
        res.message = sql_error(SQL_NULL_HENV,SQL_NULL_HDBC,hstmt);
        SQLFreeStmt(hstmt,SQL_DROP);
        return res;
    }

    SQLFreeStmt(hstmt,SQL_DROP);

    return res;
}

void OdbcServer::split_name( const char * source, SQLCHAR *catalog, 
        SQLCHAR *schema, SQLCHAR *table )
{
    char *ptr1, *ptr2;
    char working[ 256 ];

    strcpy( working, source );

    ptr1 = ptr2 = NULL;

    ptr1 = strchr( source, '.' );
    if ( ptr1 )
    {
        *ptr1 = '\0';
        source = ptr1 + 1;
        ptr2 = strchr( source, '.' );

        if ( ptr2 )
        {
            *ptr2 = '\0';
            source = ptr2 + 1;
        }
    }

    if ( ptr1 )
    {
        strcpy(( char * ) catalog, working );
    }
    else
    {
        strcpy(( char * ) catalog, "" );
    }
    if ( ptr2 )
    {
        strcpy(( char * ) schema, ptr1 + 1 );
    }
    else
    {
        strcpy(( char * ) schema, "" );
    }
    strcpy(( char * ) table, source );
}

QueryResult OdbcServer::getColumns(QString baseName, QString tableName)
{
    QueryResult res;
    SQLCHAR *pschema;
    SQLCHAR *ptables;
    SQLCHAR *pcatalog;
    SQLRETURN ret;
    SQLCHAR catalog[ 128 ], schema[ 128 ], table[ 128 ];
    SQLHANDLE hstmt;

    if ( tableName.isNull() )
    {
      ptables = NULL;
      pschema = NULL;
      pcatalog = NULL;
    }
    else
    {
        split_name( tableName, catalog, schema, table );
        ptables = table;
        pschema = schema;
        pcatalog = catalog;
    }

    res.type = QueryResult::Set;

    // obtain a list of schema

    ret = SQLAllocStmt( hdbc, &hstmt );
    if ( !SQL_SUCCEEDED( ret ))
    {
        res.type = QueryResult::Error;
        res.message = sql_error(SQL_NULL_HENV,hdbc,SQL_NULL_HSTMT);
        return res;
    }
    
    ret = SQLColumns( hstmt, 
            pcatalog, SQL_NTS, 
            pschema, SQL_NTS, 
            ptables, SQL_NTS, 
            (SQLCHAR*)"", 0 );

    if ( !SQL_SUCCEEDED( ret ))
    {
        res.type = QueryResult::Error;
        res.message = sql_error(SQL_NULL_HENV,SQL_NULL_HDBC,hstmt);
        SQLFreeStmt(hstmt,SQL_DROP);
        return res;
    }

    res.type = QueryResult::Set;

    while( SQL_SUCCEEDED(( ret = SQLFetch( hstmt ))))
    {
        char schema[ 128 ];
        SQLINTEGER len;
        SQLRETURN ret1;

        ret1 = SQLGetData( hstmt, 4, SQL_C_CHAR, schema, sizeof( schema ),
                &len );

        if ( SQL_SUCCEEDED( ret1 ) && len != SQL_NULL_DATA )
        {
  	        res.result.append( schema );
        }
    }

    if ( ret != SQL_NO_DATA )
    {
        res.type = QueryResult::Error;
        res.message = sql_error(SQL_NULL_HENV,SQL_NULL_HDBC,hstmt);
        SQLFreeStmt(hstmt,SQL_DROP);
        return res;
    }

    SQLFreeStmt(hstmt,SQL_DROP);

    return res;
}

QueryResult OdbcServer::shutdown(void)
{
      QueryResult queryRes;
      queryRes.type = QueryResult::Error;
      queryRes.message = "Sorry, this is not supported by the ODBC plugin ;(";
      return queryRes;
}

QueryResult OdbcServer::reload(void)
{
    QueryResult queryRes;
    queryRes.type = QueryResult::Error;
    queryRes.message = "Sorry, this is not supported by the ODBC plugin ;(";
    return queryRes;
}

QString OdbcServer::serverInfo(void)
{
    return "Type: ODBC\n";
}

QueryResult OdbcServer::createBase(QString baseName)
{
      QueryResult queryRes;
      queryRes.type = QueryResult::Error;
      queryRes.message = "Sorry, this is not supported by the ODBC plugin ;(";
      return queryRes;
}

QueryResult OdbcServer::dropBase(QString baseName)
{
      QueryResult queryRes;
      queryRes.type = QueryResult::Error;
      queryRes.message = "Sorry, this is not supported by the ODBC plugin ;(";
      return queryRes;
}

IObject *OdbcServer::queryInterface(UIID id)
{
    if(id == UIID_SQLSERVER)
        return this;
    else
        return 0;
}
